import inspect
from collections import defaultdict

from .service import Service


class ApiMethodError(Exception):
    pass


def apiclass(cls):
    cls._service_type = ApiClass
    cls._service_name = cls.__name__
    return cls


class ApiProxyObject:
    def __init__(self, service, instance_id, sync=False):
        self.service = service
        self.host = service.host
        self.id = instance_id
        self.sync = sync

    def __getattribute__(self, name):
        service = object.__getattribute__(self, 'service')
        id = object.__getattribute__(self, 'id')

        if name == '_service':
            return service
        if name == '_id':
            return id

        served_cls = service.served_cls

        try:
            served_meth = getattr(served_cls, name)
            if served_meth and callable(served_meth):
                return ApiMethod(service, name, id)
        except Exception as e:
            raise ApiMethodError(e)

        raise ValueError(
            "Served class {served_class.__name__} has no method {name}"
        )


class ApiMethod(Service):
    def __init__(self, class_service, method_name, instance_id):
        self.class_service = class_service
        self.method_name = method_name

        super().__init__(f'{class_service.name}.{method_name}')

        self.instance_id = instance_id
        self.is_remote = class_service.is_remote
        self.host = self.class_service.host
        self.host_id = self.class_service.host_id

    def task(self, *args, **kwargs):
        if self.is_remote:
            rv = self.host.call_with_cb(
                self, args, kwargs,
            )
        else:
            raise ApiMethodError("Method with callback can only be invoked remotely")

        return rv

    def cb(self, *args, callback, **kwargs):
        cls = self.class_service.served_cls
        method = getattr(cls, self.method_name)
        needs_response = getattr(method, "needs_response", True)
        if self.is_remote:
            rv = self.host.call_with_cb(
                self, args, kwargs,
                response=needs_response, response_callback=callback
            )
        else:
            raise ApiMethodError("Method with callback can only be invoked remotely")

        return rv

    def sync(self, *args, **kwargs):
        cls = self.class_service.served_cls
        method = getattr(cls, self.method_name)
        needs_response = getattr(method, "needs_response", True)
        if self.is_remote:
            rv = self.host.call_blocking(
                self, args, kwargs, response=needs_response,
            )
        else:
            raise ApiMethodError("Method sync-call can only be invoked remotely")

        return rv

    async def __call__(self, *args, **kwargs):
        cls = self.class_service.served_cls
        method = getattr(cls, self.method_name)
        needs_response = getattr(method, "needs_response", True)
        if self.is_remote:
            try:
                rv = await self.host.call(
                    self, args, kwargs, response=needs_response,
                )
            except Exception as e:
                raise
        else:
            cls = self.class_service.served_cls
            instance = ApiClass.instance_map[self.class_service.name].get(self.instance_id)
            rv = method(instance, *args, **kwargs)
            if inspect.isawaitable(rv):
                rv = await rv
        return rv


class ApiNonInstanceMethod(Service):
    def __init__(self, class_service, method_name):
        self.class_service = class_service
        self.method_name = method_name

        super().__init__(f'{class_service.name}.{method_name}')

        self.is_remote = class_service.is_remote
        self.host = self.class_service.host
        self.host_id = self.class_service.host_id

    def sync(self, *args, **kwargs):
        cls = self.class_service.served_cls
        method = getattr(cls, self.method_name)
        needs_response = getattr(method, "needs_response", True)

        if self.is_remote:
            rv = self.host.call_sync(
                self, args, kwargs, response=needs_response,
            )
        else:
            rv = method(*args, **kwargs)

        return rv

    async def __call__(self, *args, **kwargs):
        cls = self.class_service.served_cls
        method = getattr(cls, self.method_name)
        needs_response = getattr(method, "needs_response", True)

        if self.is_remote:
            rv = await self.host.call(
                self, args, kwargs, response=needs_response,
            )
        else:
            rv = method(*args, **kwargs)
            if inspect.isawaitable(rv):
                rv = await rv
        return rv


class ApiClass(Service):
    instance_map = defaultdict(dict)
    instance_next_id = 1

    def __init__(self, served_cls):
        self.served_cls = served_cls
        self.sync = False

        name = self.served_cls.__name__
        if hasattr(served_cls, '_service_name'):
            name = served_cls._service_name

        super().__init__(name)

    async def __call__(self, *args, **kwargs):
        """
        __call__() implements the constructor, i.e. instance = ClassService()

        """
        if self.is_remote:
            instance_id = await self.host.call(self, args, kwargs)
            rv = ApiProxyObject(self, instance_id, sync=self.sync)
        else:
            instance = self.served_cls(*args, **kwargs)
            if inspect.isawaitable(instance):
                instance = await instance
            instance_id = ApiClass.instance_next_id
            ApiClass.instance_next_id += 1
            ApiClass.instance_map[self.name][instance_id] = instance
            rv = instance_id

        return rv

    def __getattr__(self, name):
        served_cls = self.served_cls

        try:
            served_meth = getattr(served_cls, name)
            if served_meth and callable(served_meth):
                return ApiNonInstanceMethod(self, name)
        except Exception as e:
            raise ApiMethodError(e)

        return AttributeError(f"{self.name} has no attribute {name}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pytypes.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rpytypes.proto\x12\x04\x63\x61rp\"\xe5\x01\n\x0bPythonValue\x12\x11\n\x07_string\x18\x01 \x01(\tH\x00\x12\x11\n\x07_double\x18\x02 \x01(\x01H\x00\x12\x0e\n\x04_int\x18\x03 \x01(\x12H\x00\x12\x10\n\x06_bytes\x18\x04 \x01(\x0cH\x00\x12\x15\n\x0b_serialized\x18\x05 \x01(\x0cH\x00\x12\x0f\n\x05_bool\x18\x06 \x01(\x08H\x00\x12\x0f\n\x05_none\x18\x07 \x01(\x08H\x00\x12#\n\x06_array\x18\x08 \x01(\x0b\x32\x11.carp.PythonArrayH\x00\x12!\n\x05_dict\x18\t \x01(\x0b\x32\x10.carp.PythonDictH\x00\x42\r\n\x0bvalue_types\"R\n\x0ePythonDictItem\x12\x1e\n\x03key\x18\x01 \x01(\x0b\x32\x11.carp.PythonValue\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.carp.PythonValue\"1\n\nPythonDict\x12#\n\x05items\x18\x01 \x03(\x0b\x32\x14.carp.PythonDictItem\"/\n\x0bPythonArray\x12 \n\x05items\x18\x01 \x03(\x0b\x32\x11.carp.PythonValue\"8\n\x14PythonValueContainer\x12 \n\x05value\x18\x01 \x01(\x0b\x32\x11.carp.PythonValue\"8\n\x14PythonArrayContainer\x12 \n\x05value\x18\x01 \x01(\x0b\x32\x11.carp.PythonArray\"6\n\x13PythonDictContainer\x12\x1f\n\x05value\x18\x01 \x01(\x0b\x32\x10.carp.PythonDictb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pytypes_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PYTHONVALUE._serialized_start=24
  _PYTHONVALUE._serialized_end=253
  _PYTHONDICTITEM._serialized_start=255
  _PYTHONDICTITEM._serialized_end=337
  _PYTHONDICT._serialized_start=339
  _PYTHONDICT._serialized_end=388
  _PYTHONARRAY._serialized_start=390
  _PYTHONARRAY._serialized_end=437
  _PYTHONVALUECONTAINER._serialized_start=439
  _PYTHONVALUECONTAINER._serialized_end=495
  _PYTHONARRAYCONTAINER._serialized_start=497
  _PYTHONARRAYCONTAINER._serialized_end=553
  _PYTHONDICTCONTAINER._serialized_start=555
  _PYTHONDICTCONTAINER._serialized_end=609
# @@protoc_insertion_point(module_scope)



from .service import Service
from .api_function import ApiFunction
from .call_data import CallData, CallResponse
from .api_class import ApiClass, apiclass, ApiProxyObject, ApiMethod, ApiNonInstanceMethod

def noresp(method):
    method.needs_response = False
    return method






import json
import datetime
from dateutil import parser
from .serializer import Serializer
from .serializable import Serializable


class CarpJSONEncoder(json.JSONEncoder):
    def default(self, obj): 
        from .serializable import Serializable
        if isinstance(obj, datetime.datetime):
            return dict(__datetime__=True, value=obj.isoformat())
        elif isinstance(obj, datetime.date):
            return dict(__date__=True, value=obj.isoformat())
        elif isinstance(obj, Serializable):
            return dict(__serializable__=True, value=obj.serialize())
        return json.JSONEncoder.default(self, obj)


def decode_hook(obj):
    if isinstance(obj, dict) and "__datetime__" in obj:
        return parser.isoparse(obj["value"])
    elif isinstance(obj, dict) and "__date__" in obj:
        return parser.isoparse(obj["value"]).date()
    elif isinstance(obj, dict) and "__serializable__" in obj:
        return Serializer.deserialize(obj["value"])
    return obj


class JsonSerializer(Serializer):
    label = "json"

    def __init__(self, **serialize_options):
        self.serialize_options = serialize_options
        super().__init__()

    def serialize(self, pyobj):
        if isinstance(pyobj, dict):
            obj_dict = pyobj
            obj_dict['__type__'] = type(pyobj).__name__
        elif hasattr(pyobj, 'to_dict'):
            obj_dict = pyobj.to_dict()
            obj_dict['__type__'] = type(pyobj).__name__
        else:
            obj_dict = dict(__value__=pyobj)

        payload = json.dumps(
            obj_dict,
            **self.serialize_options,
            cls=CarpJSONEncoder
        )
        return f"{self.label}:{payload}".encode('utf-8')

    def deserialize(self, bytestr):
        obj_dict = json.loads(bytestr.decode(), object_hook=decode_hook)
        if '__type__' in obj_dict:
            typename = obj_dict['__type__']
            del obj_dict['__type__']

            if typename == 'dict':
                del obj_dict['__type__']
                return obj_dict

            typeobj = Serializable.registry.get(typename)
            if not typeobj:
                raise ValueError(f"[deserialize] Type obj not found for {typename}")
            if hasattr(typeobj, 'from_dict'):
                return typeobj.from_dict(obj_dict)
            else:
                return typeobj(**obj_dict)

        elif '__value__' in obj_dict:
            return obj_dict['__value__']
        else:
            return obj_dict
